# * Event. This represents a "goal" for the player. An event can be waited
#   on. Events can also have "meanwhile" events. If A is a meanwhile event of
#   event B, A's lifetime is a subset of B's lifetime. If B ends, A ends
#   immediately.
import pygame.constants
from data import Characters
from pygame.rect import Rect
from Vector import listify

# Events:
class Event(object):
    def __init__(self, meanwhile=None):
        self.meanwhile = meanwhile
        
    def check(self, game):
        if self.meanwhile and self.meanwhile.check(game):
            self.meanwhile.stop(game)
            self.meanwhile = None
        return self._check(game)
    def _check(self, game):
        pass

    def handle_key(self, key):
        pass
        
    def value(self, game):
        return None

    def start(self, game):
        game.goals.append(self)
        if self.meanwhile: self.meanwhile.start(game)
        self._start(game)
    def _start(self, game):
        pass

    def stop(self, game):
        game.goals.remove(self)
        if self.meanwhile: self.meanwhile.stop(game)
        self._stop(game)
    def _stop(self, game):
        pass

class ObjectIn(Event):
    def __init__(self, shape, area, rel=None, **kwargs):
        super(ObjectIn, self).__init__(**kwargs)
        self.shape = shape
        self.rel = rel
        self.area = Rect(area)

    def _check(self, game):
        if self.rel:
            area = self.area.move(game.geometry.get_pos(self.rel._object))
        else: area = self.area
        return game.geometry.contained(self.shape._object, area)

class Kills(Event):
    def __init__(self, shape, **kwargs):
        super(Kills, self).__init__(**kwargs)
        self.obj = shape._object

    def _check(self, game):
        if isinstance(self.obj.health, int):
            return self.obj.health <= 0
        return False # health = None or some similar nonsense

class Isnt(Event):
    def __init__(self, event, **kwargs):
        super(Isnt, self).__init__(**kwargs)
        self.event = event

    def _check(self, game):
        return self.event.check(game)

    def _start(self, game):
        self.event.start(game)

    def _stop(self, game):
        self.event.stop(game)

class Which(Event):
    def __init__(self, *events, **kwargs):
        super(Which, self).__init__(**kwargs)
        self.events = listify(events)

    def _check(self, game):
        for i, e in enumerate(self.events):
            if e.check(game):
                self.success = i
                return True

    def value(self, game):
        return self.success

    def _start(self, game):
        for e in self.events:
            e.start(game)

    def _stop(self, game):
        for e in self.events:
            e.stop(game)

class All(Event):
    def __init__(self, *events, **kwargs):
        super(All, self).__init__(**kwargs)
        self.events = listify(events)

    def _check(self, game):
        for e in self.events:
            if not e.check(game):
                return False
        return True

    def _start(self, game):
        for e in self.events:
            e.start(game)

    def _stop(self, game):
        for e in self.events:
            e.stop(game)

class Sequence(Event):
    '''Sequencing meta-event.

    Create with a list of events. This event will start the first, check it
    until finished, and then stop it before proceeding to the next, and
    repeating.'''
    def __init__(self, *events, **kwargs):
        super(Sequence, self).__init__(**kwargs)
        if isinstance(events[0], list):
            raise TypeError, "events are passed as *args"
        self.events = list(events)
        self.n = 0

    def _check(self, game):
        # This is a hack -- more elegant away to do simple things?
        while self.n < len(self.events) and self.events[self.n].check(game):
            self.events[self.n].stop(game)
            self.n += 1
            if self.n < len(self.events): # Events left?
                self.events[self.n].start(game) # start next
        return self.n == len(self.events)

    def _start(self, game):
        self.events[self.n].start(game)

    def _stop(self, game):
        if self.n < len(self.events):
            self.events[self.n].stop(game)

class Seconds(Event):
    def __init__(self, n, **kwargs):
        super(Seconds, self).__init__(**kwargs)
        self.n = n
    def _start(self, game):
        self.t0 = pygame.time.get_ticks()

    def _check(self, game):
        return pygame.time.get_ticks() - self.t0 > self.n*1000

class Ticks(Event):
    def __init__(self, n, **kwargs):
        super(Ticks, self).__init__(**kwargs)
        self.n = n

    def _start(self, game):
        self.target = game.frames + self.n

    def _check(self, game):
        return game.frames >= self.target

class NoneEvent(Event):
    def __init__(self, **kwargs):
        super(NoneEvent, self).__init__(**kwargs)

    def _check(self, game):
        return True

Always = NoneEvent
class Never(Event):
    def __init__(self, **kwargs):
        super(Never, self).__init__(**kwargs)
    def _check(self, game):
        return False

class Dialog(Event):
    def __init__(self, text, character=None, forever=False, **kwargs):
        super(Dialog, self).__init__(**kwargs)
        self.text = text
        if not forever:
            self.event = Seconds(len(text)/8)
        else:
            self.event = Never()
        if character == None:
            character = Characters.Narrator
        self.character = character
        self.done = False

    def handle_key(self, key):
        if key == pygame.constants.K_RETURN:
            self.done = True
            return True

    def _check(self, game):
        if self.done: return True
        return self.event.check(game)

    def _start(self, game):
        self.event.start(game)

    def _stop(self, game):
        self.event.stop(game)

def RestartDialog(text):
    return Dialog(text, forever=True)
    
class Flag(Event):
    '''Acceptable forms:

    Flag("x") # wait until flag x is true
    Flag("x", 4) # wait until flag x is 4'''
    def __init__(self, flag, *args, **kwargs):
        if kwargs.has_key('value'):
            self.wanted = kwargs['value']
            del kwargs['value']
        if args:
            self.wanted = args[0]
        super(Flag, self).__init__(**kwargs)
        self.flag = flag

    def _check(self, game):
        if not game.flags.has_key(self.flag):
            return False

        if hasattr(self, 'wanted'):
            return game.flags[self.flag] == self.wanted
        else: return bool(game.flags[self.flag])

    def value(self, game):
        return game.flags[self.flag]
        
